(* This transformation maps [PP] programs to [UPP] programs.
   The key changes are as follows.

   First, types are dropped, so, for instance, Booleans, integers, and
   pointers are no longer distinguished.

   Second, accesses to local and global variables are
   distinguished.

   Third, operations on arrays are replaced with more elementary
   memory load and store operations.

   Fourth, instructions are inserted at the beginning of every
   procedure to initialize all local variables to zero.

   Last, the set of operators changes to better reflect that of the
   MIPS. Some instruction selection is performed to recognize and
   adequately translate certain patterns. Instruction selection takes
   the form of a set of rewriting rules, implemented inside
   [Upp2upp]. *)

open Integer
open Primitive
open MIPSOps

(* ------------------------------------------------------------------------- *)

(* A global environment maps global variables to the offset that they
   have been assigned in the global storage area. *)

type genv =
    offset StringMap.t

(* A local environment is a set of local variables. A variable is
   considered local if its name appears in the local environment,
   and global otherwise. *)

type env =
    StringSet.t

(* ------------------------------------------------------------------------- *)

(* Translating expressions. *)

let rec translate_expression (genv : genv) (env : env) = function

    (* Constants. Drop type distinctions. *)

  | PP.EConst (PP.ConstBool false) ->
      UPP.EConst 0l
  | PP.EConst (PP.ConstBool true) ->
      UPP.EConst 1l
  | PP.EConst (PP.ConstInt i) ->
      UPP.EConst i

    (* Variable access. VALID ONLY FOR LOCAL VARIABLES. *)

  | PP.EGetVar v -> 
      UPP.EGetVar v

    (* Operator applications. Their sub-expressions are easily
       translated using recursive calls to
       [translate_expression]. Then, the translation of the operator
       application itself is delegated by [Upp2upp]. *)

  | PP.EUnOp (op, e) ->
      Upp2upp.mkunop op (translate_expression genv env e)

  | PP.EBinOp (op, e1, e2) ->
      Upp2upp.mkbinop op (translate_expression genv env e1) (translate_expression genv env e2)

    (* Function call. *)

  | PP.EFunCall (callee, es) ->
      UPP.EFunCall (callee, List.map (translate_expression genv env) es)

    (* Array read. We compute the element's address and access it. *)

  | PP.EArrayGet (earray, eindex) -> failwith "Pas implment !"

    (* Array allocation. Forget about the type and convert the
       desired size, which is expressed as a number of elements,
       into a number of bytes. Convert the expression into a call
       to the [Alloc] primitive function. *)

  | PP.EArrayAlloc (_, e) -> failwith "Pas implment !"

(* ------------------------------------------------------------------------- *)

(* Translating conditions. *)

let rec translate_condition (genv : genv) (env : env) = function
  | PP.CExpression e ->
      UPP.CExpression (translate_expression genv env e)
  | PP.CNot c ->
      UPP.CNot (translate_condition genv env c)
  | PP.CAnd (c1, c2) ->
      UPP.CAnd (translate_condition genv env c1, translate_condition genv env c2)
  | PP.COr (c1, c2) ->
      UPP.COr (translate_condition genv env c1, translate_condition genv env c2)

(* ------------------------------------------------------------------------- *)

(* Translating instructions. *)

let rec translate_instruction (genv : genv) (env : env) = function

    (* Procedure call. *)

  | PP.IProcCall (callee, es) -> failwith "Pas implment !"

    (* Variable update. VALID ONLY FOR LOCAL VARIABLES. *)

  | PP.ISetVar (v, e) ->
      UPP.ISetVar (v, translate_expression genv env e)

    (* Array write. We compute the element's address and access it. *)

  | PP.IArraySet (earray, eindex, evalue) -> failwith "Pas implment !"

    (* Sequence. *)

  | PP.ISeq is ->
      UPP.ISeq (List.map (translate_instruction genv env) is)

    (* Conditional. *)

  | PP.IIf (c, i1, i2) ->
      UPP.IIf (
        translate_condition genv env c,
        translate_instruction genv env i1,
        translate_instruction genv env i2
      )

    (* While. *)

  | PP.IWhile (c, i) ->
      UPP.IWhile (
        translate_condition genv env c,
        translate_instruction genv env i
      )

(* ------------------------------------------------------------------------- *)

(* Translating procedures. This involves building the procedure's
   local environment -- here, simply a set of local variable names
   that allows distinguishing global and local variables. (Local
   variables hide global variables, so [genv] alone isn't enough
   to distinguish.) *)

let translate_procedure genv f proc =

  (* Obtain a list of the formal parameters' names (without their
     types). *)

  let formals = List.map fst proc.PP.formals in

  (* Build the procedure's local environment -- here, simply a set of
    local variable names that allows distinguishing global and local
    variables. (Local variables hide global variables, so [genv] alone
    isn't enough to distinguish.) *)

  let env = StringSet.of_list formals in
  let result, env =
    match proc.PP.result with
    | None ->
	false, env
    | Some _ ->
	true, StringSet.add f env
  in
  let locals = StringMap.domain proc.PP.locals in
  let env = StringSet.union locals env in

  (* Translate the procedure's body. *)

  let body =
    [ translate_instruction genv env proc.PP.body ]
  in

  (* Insert instructions to initialize every local variable (including
     the result variable, if there is one) to zero. This is required in
     order to preserve the semantics of [PP]. Indeed, the semantics of
     [PP] states that every local variable receives a default value when
     it is allocated, whereas the semantics of [UPP] states that every
     local variable is initially undefined.

     Many of these initialization instructions are usually superfluous,
     because programmers often explicitly initialize variables before
     using them. In that case, the newly added instructions are in fact
     dead. This fact will be detected by the liveness analysis, and the
     dead instructions will be removed in the translation of [ERTL] to
     [LTL]. *)

  let initialize local =
   UPP.ISetVar (local, UPP.EConst 0l)
   in

  let body =
    if result then initialize f :: body else body
  in

  let body =
    StringMap.fold (fun local (_ : PP.typ) body ->
      initialize local :: body
    ) proc.PP.locals body
  in

  (* This is it. *)

  {
    UPP.formals = formals;
    UPP.result = result;
    UPP.locals = locals;
    UPP.body = UPP.ISeq body
  }

(* Building the global environment. EMPTY IMPLEMENTATION. *)

let allocate_globals (p : PP.program) : genv * int32 =
  (StringMap.empty, 0l)

(* Translating programs. *)

let translate_program (p : PP.program) : UPP.program =

  (* Build the global environment and compute how much space
     is needed to hold the global variables. *)

  let genv, globals = allocate_globals p in

  (* Translate the program body to a procedure, for greater
     uniformity. *)

  let main = {
    UPP.formals = [];
    UPP.result = false;
    UPP.locals = StringSet.empty;
    UPP.body = translate_instruction genv StringSet.empty p.PP.main
  } in

  (* Build a transformed program. *)

  {
    UPP.globals = globals;
    UPP.defs = StringMap.add "_main" main (StringMap.mapi (translate_procedure genv) p.PP.defs)
  }

